from PyQt5.QtWidgets import QPushButton
from PyQt5.QtGui import QIcon
import traceback

class PushButton(QPushButton):

    def __init__(self, element):
        super().__init__("", element.parent.widget)

        self.element = element

        self.clicked.connect(self.callback)

        if element.icon:
            self.setIcon(QIcon(element.icon))

        if element.text:
            if type(element.text) == str:
                txt = element.text
            else:
                txt = self.element.getvar(element.text.variable_group, element.text.variable)    
            self.setText(txt)    

        if self.element.tooltip:
            if type(self.element.tooltip) == str:
                txt = self.element.tooltip
            else:
                txt = self.element.getvar(self.element.tooltip.variable_group, self.element.tooltip.variable)    
            self.setToolTip(txt)

        self.set_geometry()

    def set(self):
        if type(self.element.text) != str:
            self.setText(self.element.getvar(self.element.text.variable_group, self.element.text.variable))   
        if type(self.element.tooltip) != str:
            self.setToolTip(self.element.getvar(self.element.tooltip.variable_group, self.element.tooltip.variable))


    def callback(self):
        try:
            if self.element.callback and self.underMouse():
                self.element.callback(self)
                # Update GUI
                self.element.window.update()
        except:
            traceback.print_exc()

    def set_geometry(self):
        x0, y0, wdt, hgt = self.element.get_position()
        self.setGeometry(x0, y0, wdt, hgt)



from .layer import Layer
from geopandas import GeoDataFrame
from pyogrio import read_dataframe


class ChoroplethLayer(Layer):
    def __init__(self, mapbox, id, map_id, **kwargs):
        super().__init__(mapbox, id, map_id, **kwargs)
        pass

    def set_data(
        self, data, color_by_attribute: dict = None, legend_items: list = None, 
    ):
        # Make sure this is not an empty GeoDataFrame
        if isinstance(data, GeoDataFrame):
            # Data is GeoDataFrame
            if len(data) == 0:
                data = GeoDataFrame()
            if data.crs != 4326:
                data = data.to_crs(4326)
        else:
            # Read geodataframe from shape file
            data = read_dataframe(data)

        self.data = data
        # self.visible = True

        if not self.big_data:
            if color_by_attribute is None:
                # Add new layer
                self.mapbox.runjs(
                    "./js/choropleth_layer.js",
                    "addLayer",
                    arglist=[
                        self.map_id,
                        self.data,
                        self.min_zoom,
                        self.hover_property,
                        self.color_property,
                        self.line_color,
                        self.line_width,
                        self.line_opacity,
                        self.fill_opacity,
                        self.legend_title,
                        self.unit,
                        self.legend_position,
                        self.side,
                        self.bins,
                        self.colors,
                        self.color_labels,
                    ],
                )
            elif isinstance(color_by_attribute, dict) and isinstance(
                legend_items, list
            ):
                # Color by attribute
                self.mapbox.runjs(
                    "./js/polygon_layer_custom.js",
                    "addLayer",
                    arglist=[
                        self.map_id,
                        self.data,
                        self.hover_property,
                        self.min_zoom,
                        color_by_attribute,
                        legend_items,
                        self.legend_position,
                        self.legend_title,
                    ],
                )
            elif isinstance(color_by_attribute, dict
            ):
                # Color by attribute
                self.mapbox.runjs(
                    "./js/polygon_layer_additional_attr.js",
                    "addLayer",
                    arglist=[
                        self.map_id,
                        self.data,
                        self.hover_property,
                        self.min_zoom,
                        color_by_attribute,
                        legend_items,
                        self.legend_position,
                        self.legend_title,
                    ],
                )
        else:
            self.update()

    def update(self):
        if self.data is None:
            return
        if len(self.data) == 0:
            # Empty GeoDataFrame
            return
        # Only need to update this layer if it use big data and is visible
        if self.big_data and self.get_visibility():
            if self.mapbox.zoom > self.min_zoom:
                # Zoomed in
                coords = self.mapbox.map_extent
                xl0 = coords[0][0]
                xl1 = coords[1][0]
                yl0 = coords[0][1]
                yl1 = coords[1][1]
                # Limits WGS 84
                gdf = self.data.cx[xl0:xl1, yl0:yl1]
                # Remove existing layer
                # Add new layer
                self.mapbox.runjs(
                    "./js/choropleth_layer.js",
                    "addLayer",
                    arglist=[
                        self.map_id,
                        gdf,
                        self.min_zoom,
                        self.hover_property,
                        self.color_property,
                        self.line_color,
                        self.line_width,
                        self.line_opacity,
                        self.fill_opacity,
                        self.legend_title,
                        self.unit,
                        self.legend_position,
                        self.side,
                        self.bins,
                        self.colors,
                        self.color_labels,
                    ],
                )
            else:
                # Zoomed out
                # Choropleths are automatically invisible, but legend is not
                self.mapbox.runjs(self.main_js, "hideLegend", arglist=[self.map_id])

    def activate(self):
        self.active = True
        self.mapbox.runjs(
            "./js/choropleth_layer.js",
            "activate",
            arglist=[
                self.map_id,
                self.line_color,
                self.fill_color,
                self.line_color_selected,
                self.fill_color_selected,
            ],
        )

    def deactivate(self):
        self.active = False
        self.mapbox.runjs(
            "./js/choropleth_layer.js",
            "deactivate",
            arglist=[
                self.map_id,
                self.line_color_inactive,
                self.line_width_inactive,
                self.line_style_inactive,
                self.line_opacity_inactive,
                self.fill_color_inactive,
                self.fill_opacity_inactive,
                self.line_color_selected_inactive,
                self.fill_color_selected_inactive,
            ],
        )

    def redraw(self):
        if isinstance(self.data, GeoDataFrame):
            self.set_data(self.data)
        if not self.get_visibility():
            self.set_visibility(False)



import os

import geojson

from .layer import Layer
from geopandas import GeoDataFrame
import matplotlib.colors as mcolors

class LineLayer(Layer):
    def __init__(self, mapbox, id, map_id, **kwargs):
        super().__init__(mapbox, id, map_id, **kwargs)

        pass


    def set_data(self,
                 data):

        self.data = data   

        # Make sure this is not an empty GeoDataFrame
        if isinstance(data, GeoDataFrame):
            # Data is GeoDataFrame
            if len(data) == 0:
                return
        else:
            print("Data is not a GeoDataFrame")
            return    

        # Add new layer        
        self.mapbox.runjs("./js/line_layer.js", "addLayer", arglist=[self.map_id,
                                                                     data.to_crs(4326),
                                                                     self.line_color,
                                                                     self.line_width,
                                                                     self.line_opacity,
                                                                     self.fill_color,
                                                                     self.fill_opacity,
                                                                     self.circle_radius])
    def redraw(self):
        if isinstance(self.data, GeoDataFrame):
            self.set_data(self.data)
        if not self.get_visibility():
            self.set_visibility(False)

    def activate(self):
        self.mapbox.runjs("./js/line_layer.js", "setPaintProperties", arglist=[self.map_id,
                                                                               self.line_color,
                                                                               self.line_width,
                                                                               self.line_opacity,
                                                                               self.fill_color,
                                                                               self.fill_opacity,
                                                                               self.circle_radius])
  
    def deactivate(self):
        self.mapbox.runjs("./js/line_layer.js", "setPaintProperties", arglist=[self.map_id,
                                                                               self.line_color_inactive,
                                                                               self.line_width_inactive,
                                                                               self.line_opacity_inactive,
                                                                               self.fill_color_inactive,
                                                                               self.fill_opacity_inactive,
                                                                               self.circle_radius_inactive])

    # def set_visibility(self, true_or_false):
    #     if true_or_false:
    #         self.mapbox.runjs("/js/main.js", "showLayer", arglist=[self.map_id + ".line"])
    #         self.mapbox.runjs("/js/main.js", "showLayer", arglist=[self.map_id + ".circle"])
    #     else:
    #         self.mapbox.runjs("/js/main.js", "hideLayer", arglist=[self.map_id + ".line"])
    #         self.mapbox.runjs("/js/main.js", "hideLayer", arglist=[self.map_id + ".circle"])




import os

import geojson

from .layer import Layer
from geopandas import GeoDataFrame
import matplotlib.colors as mcolors

class LineSelectorLayer(Layer):
    def __init__(self, mapbox, id, map_id, **kwargs):
        super().__init__(mapbox, id, map_id, **kwargs)

        pass


    def set_data(self,
                 data,
                 index):

        self.data = data
        self.index = index

        # Make sure this is not an empty GeoDataFrame
        if isinstance(data, GeoDataFrame):
            # Data is GeoDataFrame
            if len(data) == 0:
                data = GeoDataFrame()
            # Make sure there is an index column
            data["index"] = range(len(data))

        # Add new layer        
        self.mapbox.runjs("./js/line_selector_layer.js", "addLayer", arglist=[self.map_id,
                                                                               data.to_crs(4326),
                                                                               index,
                                                                               self.line_color,
                                                                               self.line_width,
                                                                               self.line_style,
                                                                               self.line_opacity,
                                                                               self.line_color_selected,
                                                                               self.line_width_selected,
                                                                               self.line_style_selected,
                                                                               self.line_opacity_selected,
                                                                               self.hover_param,
                                                                               self.selection_type])

        # if self.mode == "inactive":
        #     self.deactivate()
        if not self.active:
            self.deactivate()

    def set_selected_index(self, index):
        self.mapbox.runjs("/js/line_selector_layer.js", "setSelectedIndex", arglist=[self.map_id, index])


    def activate(self):
        # The line_selector_layer has no setPaintProperties yet.
        return
        self.mapbox.runjs("./js/line_selector_layer.js", "setPaintProperties", arglist=[self.map_id,
                                                                               self.line_color,
                                                                               self.line_width,
                                                                               self.line_style,
                                                                               self.line_opacity,
                                                                               self.line_color_selected,
                                                                               self.line_width_selected,
                                                                               self.line_style_selected,
                                                                               self.line_opacity_selected])
  
    def deactivate(self):
        # The line_selector_layer has no setPaintProperties yet.
        return
        self.mapbox.runjs("./js/line_selector_layer.js", "setPaintProperties", arglist=[self.map_id,
                                                                                         self.line_color_inactive,
                                                                                         self.line_width_inactive,
                                                                                         self.line_style_inactive,
                                                                                         self.line_opacity_inactive])

    # def set_visibility(self, true_or_false):
    #     if true_or_false:
    #         self.mapbox.runjs("/js/main.js", "showLayer", arglist=[self.map_id + ".line"])
    #         self.mapbox.runjs("/js/main.js", "showLayer", arglist=[self.map_id + ".circle"])
    #     else:
    #         self.mapbox.runjs("/js/main.js", "hideLayer", arglist=[self.map_id + ".line"])
    #         self.mapbox.runjs("/js/main.js", "hideLayer", arglist=[self.map_id + ".circle"])

    def redraw(self):
        if isinstance(self.data, GeoDataFrame):
            self.set_data(self.data, self.index)
        if not self.get_visibility():
            self.set_visibility(False)



from PyQt5.QtWidgets import QTableView
from PyQt5.QtCore import QItemSelection, QItemSelectionModel
from PyQt5.QtWidgets import QLabel
from PyQt5 import QtCore
import traceback

from enum import Enum

import pandas as pd
import copy

class DataFrameModel(QtCore.QAbstractTableModel):
    """Class to create table view from dataframe"""

    DtypeRole = QtCore.Qt.UserRole + 1000
    ValueRole = QtCore.Qt.UserRole + 1001

    def __init__(self, df=pd.DataFrame(), parent=None):
        super(DataFrameModel, self).__init__(parent)
        self._dataframe = df

    def setDataFrame(self, dataframe):
        self.beginResetModel()
        self._dataframe = dataframe.copy()
        self.endResetModel()

    def dataFrame(self):
        return self._dataframe

    dataFrame = QtCore.pyqtProperty(pd.DataFrame, fget=dataFrame, fset=setDataFrame)

    @QtCore.pyqtSlot(int, QtCore.Qt.Orientation, result=str)
    def headerData(
        self,
        section: int,
        orientation: QtCore.Qt.Orientation,
        role: int = QtCore.Qt.DisplayRole,
    ):
        if role == QtCore.Qt.DisplayRole:
            if orientation == QtCore.Qt.Horizontal:
                return self._dataframe.columns[section]
            else:
                return str(self._dataframe.index[section])
        return QtCore.QVariant()

    def rowCount(self, parent=QtCore.QModelIndex()):
        if parent.isValid():
            return 0
        return len(self._dataframe.index)

    def columnCount(self, parent=QtCore.QModelIndex()):
        if parent.isValid():
            return 0
        return self._dataframe.columns.size

    def data(self, index, role):
        if not index.isValid() or not (
            0 <= index.row() < self.rowCount()
            and 0 <= index.column() < self.columnCount()
        ):
            return QtCore.QVariant()
        row = self._dataframe.index[index.row()]
        col = self._dataframe.columns[index.column()]
        dt = self._dataframe[col].dtype

        val = self._dataframe.iloc[row][col]
        if role == QtCore.Qt.DisplayRole:
            return str(val)
        elif role == DataFrameModel.ValueRole:
            return val
        if role == DataFrameModel.DtypeRole:
            return dt
        return QtCore.QVariant()

    def roleNames(self):
        roles = { 
            QtCore.Qt.DisplayRole: b"display",
            DataFrameModel.DtypeRole: b"dtype",
            DataFrameModel.ValueRole: b"value",
        }
        return roles

    def sort(self, column, order):
        colname = self._dataframe.columns.tolist()[column]
        self.layoutAboutToBeChanged.emit()
        self._dataframe.sort_values(
            colname, ascending=order == QtCore.Qt.AscendingOrder, inplace=True
        )
        self._dataframe.reset_index(inplace=True, drop=True)
        self.layoutChanged.emit()


class SelectionTable(Enum):
    # NoSelection = 0
    # SingleSelection = 1
    # MultiSelection = 2
    # ExtendedSelection = 3
    none = 0
    single = 1
    multiple = 2
    extended = 3


class TableView(QTableView):
    """Table pyqt Element"""

    def __init__(self, element):
        super().__init__(element.parent.widget)
        self.element = element

        # Add title
        if element.text:
            label = QLabel(element.text, element.parent.widget)
            label.setStyleSheet("background: transparent; border: none")
            self.text_widget = label

        # Connect callback when event happens
        self.clicked.connect(self.callback)

        # This allows for whole row or column selection and not individual cells. Use row as default
        selection_direction = QTableView.SelectRows
        if hasattr(self.element, "selection_direction"):
            if element.selection_direction == "column":
                selection_direction = QTableView.SelectColumns
                if self.element.sortable:
                    raise ValueError("Column selection cannot be sortable")
        self.setSelectionBehavior(selection_direction)

        # If the header is clicked and the selection direction is rows, the table is sorted and header_clicker is called
        if self.element.sortable and selection_direction == QTableView.SelectRows:
            self.setSortingEnabled(True)
            self.horizontalHeader().sectionClicked.connect(self.header_clicked)
            # Make sure that the sorting direction is shown
            self.horizontalHeader().setSortIndicatorShown(True)
        elif selection_direction == QTableView.SelectColumns:
            # If the selection direction is columns, the table is not sorted but it does count as selecting the column. Call the regular callback function
            self.horizontalHeader().sectionClicked.connect(self.callback)
        else:    
            self.setSortingEnabled(False)

        self.set_geometry()

        # Set selection mode
        selection_type = None
        if hasattr(self.element, "selection_type"):
            selection_type = getattr(SelectionTable, element.selection_type, None)
        self.multi_selection = False
        if selection_type:
            self.setSelectionMode(selection_type.value)
            if (
                selection_type == SelectionTable.multiple
                or selection_type == SelectionTable.extended
            ):
                self.multi_selection = True

        self.execute_callback = True

        self.sort_order = 0
        self.sort_column = 0

    def resize_columns(self, df):
        for i, col in enumerate(df.columns):
            max_char = max(max([len(str(x)) for x in df[col].values]), len(col))
            self.element.widget.setColumnWidth(i, max_char * 10)

    def set(self):
        # option_value
        df = self.element.getvar(self.element.option_value.variable_group, self.element.option_value.variable)
        if df is not None:
            self.df = df # Original unsorted dataframe
            df_sorted = copy.copy(df.reset_index()) # Sorted dataframe
            df_sorted = df_sorted.drop("index", axis=1)
        else:
            df_sorted = df
        # Update items
        self.clearSpans()
        model = DataFrameModel(df_sorted)
        self.setModel(model)
        # resizeColumnsToContents can be slow for large tables
        if df_sorted.shape[0] * df_sorted.shape[1] > 1000:
            self.resize_columns(df_sorted)
        else:
            self.resizeColumnsToContents()  
        self.verticalHeader().setVisible(False)

        if self.element.sortable:
            self.sortByColumn(self.sort_column, self.sort_order)

        # In case no selection is used skip
        if not self.element.variable:
            return

        # Find indices in sorted dataframe
        indices = self.find_sorted_indices()
        if self.selectionBehavior() == QTableView.SelectRows:
            self.select_rows(indices)
        elif self.selectionBehavior() == QTableView.SelectColumns:
            self.select_cols(indices)
        else:
            raise ValueError("Selection behavior not recognized")
        self.execute_callback = True

    def callback(self):
        # Find selected indices in the original dataframe
        if not self.element.variable:
            return
        if self.selectionBehavior() == QTableView.SelectRows:
            indices = self.find_original_indices()
        elif self.selectionBehavior() == QTableView.SelectColumns:
            # Column selection cannot be sorted
            indices = [index.column() for index in self.selectionModel().selectedColumns()]
        else:
            raise ValueError("Selection behavior not recognized")
        # Set the new value 
        name = self.element.variable
        group = self.element.variable_group
        self.element.setvar(group, name, indices)
        try:
            # Execute callback
            if self.isEnabled() and self.element.callback:
                self.element.callback(indices, self)
            # Update GUI
            self.element.window.update()
        except:
            traceback.print_exc()

    def header_clicked(self, column_index):
        """Callback when header is clicked to sort"""
        if not self.element.variable:
            return
        self.sort_column = column_index
        self.sort_order = self.horizontalHeader().sortIndicatorOrder()
        indices = self.find_sorted_indices()
        self.select_rows(indices)

    def select_rows(self, indices):
        # Select rows in table
        self.selectionModel().clearSelection()
        model = self.model() # get data model for indexes.
        selection = QItemSelection()
        for i in indices:
            # Get the model index for selection.
            # Column shouldn't matter for row-wise.
            model_index = model.index(i, 0)
            # Select single row.
            selection.select(model_index, model_index)  # top left, bottom right identical
        mode = QItemSelectionModel.Select | QtCore.QItemSelectionModel.Rows
        # Apply the selection, using the row-wise mode.
        self.selectionModel().select(selection, mode)

    def select_cols(self, indices):
        # Select rows in table
        self.selectionModel().clearSelection()
        model = self.model() # get data model for indexes.
        selection = QItemSelection()
        for i in indices:
            # Get the model index for selection.
            # Column shouldn't matter for row-wise.
            model_index = model.index(0, i)
            # Select single row.
            selection.select(model_index, model_index)  # top left, bottom right identical
        mode = QItemSelectionModel.Select | QtCore.QItemSelectionModel.Columns
        # Apply the selection, using the row-wise mode.
        self.selectionModel().select(selection, mode)

    def find_original_indices(self):
        # Find indices of row in original dataframe df0 that match the row in the sorted dataframe df
        indices = [index.row() for index in self.selectionModel().selectedRows()]
        df0 = self.df
        df = self.model()._dataframe
        for index in indices:
            row = df.iloc[index]
            index0 = df0.index[df0.apply(lambda r: r.equals(row), axis=1)].tolist()[0]
            indices[indices.index(index)] = index0
        return indices
        
    def find_sorted_indices(self):
        # Find indices of row in sorted dataframe df0 that match the row in the original dataframe df
        indices0 = self.element.getvar(self.element.variable_group, self.element.variable)
        df0 = self.df
        df = self.model()._dataframe
        indices = []
        for index in indices0:
            row = df0.loc[index]
            index0 = df.index[df.apply(lambda r: r.equals(row), axis=1)].tolist()[0]
            indices.append(index0)
        return indices

    def set_geometry(self):
        resize_factor = self.element.gui.resize_factor
        x0, y0, wdt, hgt = self.element.get_position()
        self.setGeometry(x0, y0, wdt, hgt)
        if self.element.text:
            label = self.text_widget
            fm = label.fontMetrics()
            wlab = int(fm.size(0, self.element.text).width())
            if (
                self.element.text_position == "above-center"
                or self.element.text_position == "above"
            ):
                label.setAlignment(QtCore.Qt.AlignCenter)
                label.setGeometry(
                    x0, int(y0 - 20 * resize_factor), wdt, int(20 * resize_factor)
                )
            elif self.element.text_position == "above-left":
                label.setAlignment(QtCore.Qt.AlignLeft)
                label.setGeometry(
                    x0, int(y0 - 20 * resize_factor), wlab, int(20 * resize_factor)
                )
            else:
                # Assuming left
                label.setAlignment(QtCore.Qt.AlignRight)
                label.setGeometry(
                    int(x0 - wlab - 3 * resize_factor),
                    int(y0 + 5 * self.element.gui.resize_factor),
                    wlab,
                    int(20 * resize_factor),
                )



import os
from PyQt5.QtWidgets import QPushButton, QLabel, QFileDialog
from PyQt5.QtWidgets import QLabel
from PyQt5 import QtCore
import traceback

class PushSaveFile(QPushButton):

    def __init__(self, element):
        super().__init__("", element.parent.widget)

        self.element = element

        if element.text:
            if type(element.text) == str:
                txt = element.text
            else:
                txt = self.element.getvar(element.text.variable_group, element.text.variable)    
            self.setText(txt)    

        if self.element.tooltip:
            if type(self.element.tooltip) == str:
                txt = self.element.tooltip
            else:
                txt = self.element.getvar(self.element.tooltip.variable_group, self.element.tooltip.variable)    
            self.setToolTip(txt)

        self.setVisible(True)

        self.clicked.connect(self.callback)

        self.set_geometry()
  
    def set(self):
        if type(self.element.text) != str:
            self.setText(self.element.getvar(self.element.text.variable_group, self.element.text.variable))   
        if type(self.element.tooltip) != str:
            self.setToolTip(self.element.getvar(self.element.tooltip.variable_group, self.element.tooltip.variable))
        # group  = self.element.variable_group
        # name   = self.element.variable
        # val    = self.element.getvar(group, name)
        # self.string = str(val)
        # self.setText(str(val))
        # self.setStyleSheet("")

    def callback(self):
        self.okay = True
        group = self.element.variable_group
        name  = self.element.variable
        val   = self.element.getvar(group, name)
        if not val:
            val = os.getcwd()
        fname = QFileDialog.getSaveFileName(self, self.element.title, val, self.element.filter)
        if len(fname[0])>0:
            self.element.setvar(group, name, fname[0])
        else:
            self.okay = False

        try:
            if self.okay and self.element.callback:
                val   = fname[0]
                self.element.callback(val, self)
                # Update GUI
                self.element.window.update()
        except:
            traceback.print_exc()

    def set_geometry(self):
        x0, y0, wdt, hgt = self.element.get_position()
        self.setGeometry(x0, y0, wdt, hgt)



import matplotlib.colors as mcolors

class Layer:
    def __init__(self, mapbox, id, map_id, **kwargs):
        self.mapbox    = mapbox
        self.id        = id
        self.map_id    = map_id
        self.map_ids   = [map_id]
        self.visible   = True
        self.active    = True
        self.type      = "container"
        self.layer     = {}
        self.parent    = None
        self.data      = None
        self.index     = None
        self.select    = None
        self.crs       = 4326
        self.color_values = None
        self.color_map = "jet"
        self.color_property = "value"
        self.selection_type = "single"
        self.min_zoom = 0
        self.max_zoom = 22
        self.zoom_switch = 999
        self.decimals = 1
        self.big_data = False
        self.opacity = 0.9
        self.url = None

        # Legend
        self.legend_position = "bottom-right" # Options are "top-left", "top-right", "bottom-left", "bottom-right"
        self.legend_title = "" # The text that appears at the top of the legend (e.g. Topography)
        self.legend_label = "" # The text that appears next to the legend (e.g. elevation [m])
        self.legend_units = "" # The unit string that may appear in the individual tick labels (e.g. "m")
        self.legend_decimals = -1 # -1 means automatic

        # Raster layers
        self.color_scale_auto = True # automatically scale from min to max
        self.color_scale_cmin = -1000.0
        self.color_scale_cmax =  1000.0
        self.color_scale_symmetric = True
        self.color_scale_symmetric_side = "min"
        self.hillshading = True
        self.hillshading_exaggeration = 10.0
        self.hillshading_azimuth = 315.0
        self.hillshading_altitude = 30.0
        self.quality = "medium"
        self.scale_factor = 1.0

        # Cyclone track layer
        self.show_icons = True

        # Marker layer
        self.icon_file = None
        self.icon_size = 1.0
        self.marker_color = "blue"
        self.hover_property = None
        self.click_property = None
        self.click_popup_width = None
        self.click_popup_height = None

        # Paint properties
        # Active paint properties
        self.line_color     = "dodgerblue"
        self.line_width     = 2
        self.line_style     = "-"
        self.line_opacity   = 1.0
        self.fill_color     = "dodgerblue"
        self.fill_opacity   = 1.0
        self.circle_radius  = 4
        # Inactive paint properties
        self.line_color_inactive    = "lightgrey"
        self.line_width_inactive    = 2
        self.line_style_inactive    = "-"
        self.line_opacity_inactive  = 1.0
        self.fill_color_inactive    = "lightgrey"
        self.fill_opacity_inactive  = 0.0
        self.circle_radius_inactive = 2
        # Selected paint properties 
        self.line_color_selected    = "dodgerblue"
        self.line_width_selected    = 2
        self.line_style_selected    = "-"
        self.line_opacity_selected  = 1.0
        self.fill_color_selected    = "red"
        self.fill_opacity_selected  = 1.0
        self.circle_radius_selected = 5
        # Selected inactive paint properties
        self.line_color_selected_inactive    = "lightgrey"
        self.line_width_selected_inactive    = 2
        self.line_style_selected_inactive    = "-"
        self.line_opacity_selected_inactive  = 1.0
        self.fill_color_selected_inactive    = "lightgrey"
        self.fill_opacity_selected_inactive  = 0.0
        self.circle_radius_selected_inactive = 2

        # Determine which main.js file to use 
        if type(self.mapbox).__name__ == "MapBox":
            # Regular mapbox
            self.main_js = "/js/main.js"
            self.side    = "main"
        elif type(self.mapbox).__name__ == "MapBoxCompare":
            # Compare mapbox
            self.main_js = "/js/compare.js"
            self.side = "a"

        # Set attributes based on kwargs 
        for key, value in kwargs.items():
            setattr(self, key, value)

        # Convert colors to hex
        if self.line_color != "transparent": 
            self.line_color = mcolors.to_hex(self.line_color)
        if self.fill_color != "transparent": 
            self.fill_color = mcolors.to_hex(self.fill_color)
        if self.line_color_inactive != "transparent": 
            self.line_color_inactive = mcolors.to_hex(self.line_color_inactive)
        if self.fill_color_inactive != "transparent": 
            self.fill_color_inactive = mcolors.to_hex(self.fill_color_inactive)
        if self.line_color_selected != "transparent": 
            self.line_color_selected = mcolors.to_hex(self.line_color_selected)
        if self.fill_color_selected != "transparent": 
            self.fill_color_selected = mcolors.to_hex(self.fill_color_selected)
        if self.line_color_selected_inactive != "transparent": 
            self.line_color_selected_inactive = mcolors.to_hex(self.line_color_selected_inactive)
        if self.fill_color_selected_inactive != "transparent": 
            self.fill_color_selected_inactive = mcolors.to_hex(self.fill_color_selected_inactive)

    def add_layer(self, layer_id, type=None, **kwargs):

        if layer_id in self.layer:
            # Layer already exists
            return self.layer[layer_id]

        map_id = self.map_id + "." + layer_id

        if type is None:

            # Add containing layer
            self.layer[layer_id] = Layer(self.mapbox, layer_id, map_id)
            self.layer[layer_id].parent = self
            return self.layer[layer_id]
        
        else:

            if self.type != "container":
                print("Error! Can not add layer to layer of type : " + self.type)
                return None

            if type == "circle_selector":
                from .circle_selector_layer import CircleSelectorLayer
                self.layer[layer_id] = CircleSelectorLayer(self.mapbox, layer_id, map_id, **kwargs)

            elif type == "polygon":
                from .polygon_layer import PolygonLayer
                self.layer[layer_id] = PolygonLayer(self.mapbox, layer_id, map_id, **kwargs)	
                
            elif type == "polygon_selector":
                from .polygon_selector_layer import PolygonSelectorLayer
                self.layer[layer_id] = PolygonSelectorLayer(self.mapbox, layer_id, map_id, **kwargs)

            elif type == "line_selector":
                from .line_selector_layer import LineSelectorLayer
                self.layer[layer_id] = LineSelectorLayer(self.mapbox, layer_id, map_id, **kwargs)

            elif type == "circle":
                from .circle_layer import CircleLayer
                self.layer[layer_id] = CircleLayer(self.mapbox, layer_id, map_id, **kwargs)
            
            elif type == "line":
                from .line_layer import LineLayer
                self.layer[layer_id] = LineLayer(self.mapbox, layer_id, map_id, **kwargs)

            elif type == "choropleth":
                from .choropleth_layer import ChoroplethLayer
                self.layer[layer_id] = ChoroplethLayer(self.mapbox, layer_id, map_id, **kwargs)
            
            elif type == "heatmap":
                from .heatmap_layer import HeatmapLayer
                self.layer[layer_id] = HeatmapLayer(self.mapbox, layer_id, map_id, **kwargs)

            elif type == "draw":
                from .draw_layer import DrawLayer
                self.layer[layer_id] = DrawLayer(self.mapbox, layer_id, map_id, **kwargs)

            elif type == "image":
                from .image_layer import ImageLayer
                self.layer[layer_id] = ImageLayer(self.mapbox, layer_id, map_id, **kwargs)

            elif type == "raster":
                from .raster_layer import RasterLayer
                self.layer[layer_id] = RasterLayer(self.mapbox, layer_id, map_id, **kwargs)

            elif type == "raster_from_tiles":
                from .raster_from_tiles_layer import RasterFromTilesLayer
                self.layer[layer_id] = RasterFromTilesLayer(self.mapbox, layer_id, map_id, **kwargs)

            # elif type == "deck_geojson":
            #     from .deck_geojson_layer import DeckGeoJSONLayer
            #     self.layer[layer_id] = DeckGeoJSONLayer(self.mapbox, layer_id, map_id, **kwargs)

            # elif type == "datashader_choropleth":
            #     from .datashader_choropleth_layer import DatashaderChoroplethLayer
            #     self.layer[layer_id] = DatashaderChoroplethLayer(self.mapbox, layer_id, map_id, **kwargs)

            elif type == "cyclone_track":
                from .cyclone_track_layer import CycloneTrackLayer
                self.layer[layer_id] = CycloneTrackLayer(self.mapbox, layer_id, map_id, **kwargs)

            elif type == "marker":
                from .marker_layer import MarkerLayer
                self.layer[layer_id] = MarkerLayer(self.mapbox, layer_id, map_id, **kwargs)

            elif type == "raster_tile":
                from .raster_tile_layer import RasterTileLayer
                self.layer[layer_id] = RasterTileLayer(self.mapbox, layer_id, map_id, **kwargs)

            else:
                print("Error! Layer type " + self.type + " not recognized!")
                return None

            self.layer[layer_id].type = type
            self.layer[layer_id].parent = self

            return self.layer[layer_id]
 
    def layer_added(self):
        # This function is called after a layer is added to the map (is it always called? not right now)
        if not self.visible:
            self.hide()
        if not self.active:
            self.deactivate()

    def update(self):
        # This method is called when map zoom or pan is changed
        # For certain layers, this method is overridden in the subclass
        pass

    def delete(self):
        # Delete this layer and all nested layers from map
        if self.layer:
            # Container layer
            layers = list_layers(self.layer)
            for layer in layers:
                layer.delete()
        else:        
            self.delete_from_map()

        # Remove layer from layer dict
        if self.parent:
            self.parent.layer.pop(self.id)
        else:
            self.mapbox.layer.pop(self.id)

    def clear(self):
        # Clear this layer and all nested layers from map
        if self.layer:
            # Container layer
            layers = list_layers(self.layer)
            for layer in layers:
                layer.clear()
        else:        
            self.delete_from_map()
            self.data = None

    def delete_from_map(self):
        self.mapbox.runjs(self.main_js, "removeLayer", arglist=[self.map_id, self.side])

    def set_mode(self, mode):
        # Can only be called on data layers
        if not self.layer:
            # Data layer
            if mode == "active":
                self.show()
                self.activate()
            elif mode == "inactive":
                self.show()
                self.deactivate()
            else: # Invisible   
                self.hide()
                self.deactivate()

    def show(self):
        """Show layer on map. Any child layers that are set to visible will also be shown."""
        self.visible = True
        self.set_visibility(True)

    def hide(self):
        """Hide layer on map. Any child layers will also be hidden."""
        self.visible = False
        self.set_visibility(False)

    def set_visibility(self, true_or_false):
        # Loop down through the layer hierarchy to show or hide layers
        if self.layer:
            # Container layer
            for name in self.layer:
                if true_or_false:
                    self.layer[name].show()
                else:
                    self.layer[name].hide()
        else:
            # Data layer
            if true_or_false:
                # Child may be visible, on of parents may be invisible
                if self.get_visibility():
                    self.mapbox.runjs(self.main_js, "showLayer", arglist=[self.map_id, self.side])
                else:    
                    self.mapbox.runjs(self.main_js, "hideLayer", arglist=[self.map_id, self.side])
            else:
                self.mapbox.runjs(self.main_js, "hideLayer", arglist=[self.map_id, self.side])

    def get_visibility(self):
        # Loop up through the layer hierarchy to determine if the layer is visible
        if self.visible:
            if self.parent:
                return self.parent.get_visibility()
            else:
                return True
        else:
            return False    

    def activate(self):
        # Only called for layers that do not have a activate function in the subclass
        pass

    def deactivate(self):
        # Only called for layers that do not have a deactivate function in the subclass
        pass

    def get(self, layer_id):
        if layer_id in self.layer:
            return self.layer[layer_id]
        else:
            return None

    def redraw(self):
        # This method is called when the layers style is changed
        # For most layers, this method is overridden in the subclass
        pass

def list_layers(layer_dict, layer_type="all", layer_list=None):
    if not layer_list:
        layer_list = []
    for layer_name in layer_dict:
        layer = layer_dict[layer_name]
        if layer_type != "all":
            if layer.type == layer_type:
                layer_list.append(layer)
        else:
            layer_list.append(layer)

        if layer.layer:
            layer_list = list_layers(layer.layer,
                                     layer_list=layer_list,
                                     layer_type=layer_type)
    return layer_list

def find_layer_by_id(layer_id, layer_dict, layer_type="all", layer_list=None):
    layer_list = list_layers(layer_dict)
    for layer in layer_list:
        if layer.map_id == layer_id:
            return layer



# -*- coding: utf-8 -*-
"""
Created on Wed Apr 28 10:08:26 2021

@author: ormondt
"""

import os
import pandas as pd
import matplotlib as mpl

def read_color_maps(path_name):
    """Read all colormaps from a folder and register them (add them to matplotlib.colormaps). Return list of colormap names."""    
    # Get list of files in path
    files = os.listdir(path_name)    
    # Loop over files
    for file in files:
        if file.endswith(".txt"):
            # Name is filename without extension
            name = os.path.splitext(file)[0]
            # Read colormap
            rgb = read_colormap(os.path.join(path_name, file))
            cmap = mpl.colors.ListedColormap(rgb, name=name)
            mpl.colormaps.register(cmap=cmap)
    return mpl.pyplot.colormaps()    


def cm2png(cmap,
           file_name="colorbar.png",
           orientation="horizontal",
           vmin=0.0,
           vmax=1.0,
           legend_title="",
           legend_label="",
           units="",
           unit_string="",
           decimals=-1):

    """Create png image of colormap"""
    # Create figure
    if orientation == "horizontal":
        fig = mpl.pyplot.figure(figsize=(2.5, 1))
        ax = fig.add_axes([0.05, 0.80, 0.9, 0.15])
    else:
        fig = mpl.pyplot.figure(figsize=(1, 2.5))
        ax = fig.add_axes([0.80, 0.05, 0.15, 0.90])

    norm = mpl.colors.Normalize(vmin=vmin, vmax=vmax)
    cb = mpl.colorbar.ColorbarBase(ax,
                                   cmap=cmap,
                                   norm=norm,
                                   orientation=orientation,
                                   label=legend_label)
    cb.ax.tick_params(labelsize=6)

    # Save figure
    fig.savefig(file_name, dpi=150, bbox_inches='tight')
    mpl.pyplot.close(fig)

def read_colormap(file_name):
    df = pd.read_csv(file_name, index_col=False, header=None,
                  delim_whitespace=True, names=['r','g','b'])
    v = df.to_numpy()
    return v



import os
import yaml
import sys
import copy
import shutil

from pathlib import Path
import toml
from PyQt5.QtWidgets import QApplication
from PyQt5 import QtCore, QtGui

from guitares.window import Window
from guitares.server import start_server

import guitares.icons_rc

class GUI:
    def __init__(self, module,
                 framework="pyqt5",
                 splash_file=None,
                 stylesheet=None,
                 config_path=None,
                 config_file=None,
                 icon=None,
                 server_path=None,
                 server_port=3000,
                 server_nodejs=False,
                 js_messages=True,
                 copy_mapbox_server_folder=True,
                 icon_path=None,
                 mapbox_token_file="mapbox_token.txt"):

        self.module      = module
        self.framework   = framework
        self.splash_file = splash_file
        self.stylesheet  = stylesheet
        self.config_file = config_file
        self.config_path = config_path
        self.splash      = None
        self.icon        = icon
        self.config      = {}
        self.variables   = {}
        self.server_thread = None
        self.server_path = server_path
        self.server_port = server_port
        self.server_thread = None
        self.server_nodejs = server_nodejs
        self.js_messages = js_messages
        self.popup_window = {}
        self.popup_data   = {}
        self.resize_factor = 1.0        

        QtCore.QCoreApplication.setAttribute(QtCore.Qt.AA_ShareOpenGLContexts)
        self.qtapp = QApplication(sys.argv)

        # Show splash screen
        self.show_splash()

        if not self.config_path:
            self.config_path = os.getcwd()

        self.image_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "img")
        self.image_path = self.image_path.replace(os.sep, '/')

        if server_path:
            # Need to run http server (e.g. for MapBox)
            print("Starting http server ...")
            # Run http server in separate thread
            # Use daemon=True to make sure the server stops after the application is finished
            if copy_mapbox_server_folder:
                mpboxpth = os.path.join(os.path.dirname(os.path.abspath(__file__)), "pyqt5", "mapbox", "server")
                # Delete current server folder
                if os.path.exists(server_path):
                    shutil.rmtree(server_path)
                # Now copy over folder from mapbox
                shutil.copytree(mpboxpth, server_path)
          
            # Check if mapbox token file exists in config path or in current working directory
            token_file_name = None    
            if os.path.exists(os.path.join(self.config_path, mapbox_token_file)):
                token_file_name = os.path.join(self.config_path, mapbox_token_file)
            if os.path.exists(os.path.join(os.getcwd(), mapbox_token_file)):
                token_file_name = os.path.join(os.getcwd(), mapbox_token_file)    

            # Read mapbox token and store in js file in server path
            if token_file_name:
                fid = open(token_file_name, "r")
                mapbox_token = fid.readlines()
                fid.close()
                fid = open(os.path.join(server_path, "mapbox_token.js"), "w")
                fid.write("mapbox_token = '" + mapbox_token[0].strip() + "';")
                fid.close()

            if icon_path:
                # Copy all files in icon_path to server_path/icons
                if not os.path.exists(os.path.join(server_path, "icons")):
                    os.mkdir(os.path.join(server_path, "icons"))
                for file in os.listdir(icon_path):
                    if file.endswith(".png"):
                        shutil.copy(os.path.join(icon_path, file), os.path.join(server_path, "icons", file))

            start_server(server_path, port=server_port, node=self.server_nodejs)

    def show_splash(self):
        if self.framework == "pyqt5" and self.splash_file:
            from .pyqt5.splash import Splash
            self.splash = Splash(self.splash_file, seconds=20.0).splash

    def close_splash(self):
        if self.splash:
            self.splash.close()

    def build(self):

        app = self.qtapp

        # Set the icon
        app.setWindowIcon(QtGui.QIcon(self.icon))

        if self.stylesheet:
            app.setStyleSheet(open(os.path.join(self.config_path, self.stylesheet), "r").read())

        if self.config_file:
            self.config = self.read_gui_config(self.config_path, self.config_file)

        # Make window object
        self.window = Window(self.config, self)
        self.window.build()

        # Call on_build method after building window
        if hasattr(self.module, "on_build"):
            self.module.on_build()
            
        # # # Close splash screen before GUI is initiated

        app.exec_()

    def setvar(self, group, name, value):
        if group not in self.variables:
            self.variables[group] = {}
        if name not in self.variables[group]:
            self.variables[group][name] = {}
        self.variables[group][name]["value"] = value

    def getvar(self, group, name):
        if name is None:
            # There is no variable for the element
            return None
        if group not in self.variables:
            print("Error! Cannot get variable! GUI variable group '" + group + "' not defined!")
            return None
        elif name not in self.variables[group]:
            print("Error! Cannot get variable! GUI variable '" + name + "' not defined in group '" + group + "'!")
            return None
        return self.variables[group][name]["value"]
    
    def delvar(self, group, name):
        if group not in self.variables:
            print("Error! Cannot delete variable! GUI variable group '" + group + "' not defined!")
            return None
        elif name not in self.variables[group]:
            print("Error! Cannot delete variable! GUI variable '" + name + "' not defined in group '" + group + "'!")
            return None
        del self.variables[group][name]

    def popup(self, config, id="popup", data=None):
        # Make pop-up window
        # config needs to be file name of yml file, or configuration dict
        # Data is optional and can have any shape (e.g. dict, str, object, etc.)
        # Data will only be changed if Okay is clicked in the pop-up window
        if isinstance(config, str):
            path = os.path.dirname(config)
            file_name = os.path.basename(config)
            config = self.read_gui_config(path, file_name)
        if data:    
            self.popup_data[id] = copy.copy(data)
        else:
            self.popup_data[id] = None    
        self.popup_window[id] = Window(config, self, type="popup")
        p = self.popup_window[id].build()
        okay = False
        if p.result() == 1:
            okay = True
            data = self.popup_data[id]
        # Remove popup window and data    
        return okay, data

    def read_gui_config(self, path, file_name):
        suffix = Path(path).joinpath(file_name).suffix
        if suffix == '.yml':
            d = yaml2dict(os.path.join(path, file_name))
        elif suffix == '.toml':
            d = toml.load(os.path.join(path, file_name))
        config = {}
        config["window"] = {}
        config["toolbar"] = {}
        config["menu"] = []
        config["element"] = []
        config["statusbar"] = {}
        if "window" in d:
            config["window"] = d["window"]
        if "toolbar" in d:
            config["toolbar"] = d["toolbar"]
        if "menu" in d:
            # Recursively read menu
            config["menu"] = d["menu"]
        if "element" in d:
            # Recursively read elements
            config["element"] = self.read_gui_elements(path, file_name)
        if "statusbar" in d:
            config["statusbar"] = d["statusbar"]
        return config

    def read_gui_elements(self, path, file_name):
        # Return just the elements
        suffix = Path(path).joinpath(file_name).suffix
        if suffix == '.yml':
            d = yaml2dict(os.path.join(path, file_name))
        elif suffix == '.toml':
            d = toml.load(os.path.join(path, file_name))
        element = d["element"]
        for el in d["element"]:
            if el["style"] == "tabpanel":
                # Loop through tabs
                for tab in el["tab"]:
                    if "element" in tab:
                        if type(tab["element"]) == str:
                            # Must be a file
                            tab["element"] = self.read_gui_elements(path, tab["element"])
                    else:
                        tab["element"] = []
        return element

    def quit(self):
        QApplication.quit()

def yaml2dict(file_name):
    file = open(file_name,"r")
    dct = yaml.load(file, Loader=yaml.FullLoader)
    return dct



import os
import glob
from PIL import Image
import matplotlib
from matplotlib import cm
import matplotlib.pyplot as plt
import rasterio
import rasterio.features
from rasterio.warp import calculate_default_transform, reproject, Resampling, transform_bounds
from rasterio import MemoryFile
from rasterio.transform import Affine
import geopandas as gpd
import pandas as pd
import shapely
#import matplotlib.pyplot as plt
from matplotlib.colors import LightSource
import numpy as np

from .colorbar import ColorBar
from guitares.colormap import cm2png

from .layer import Layer

class RasterLayer(Layer):
    def __init__(self, mapbox, id, map_id):
        super().__init__(mapbox, id, map_id)
        self.active = False
        self.type   = "raster"
        self.new    = True
        self.file_name = map_id + ".png"

    def activate(self):
        self.active = True

    def deactivate(self):
        self.active = False


    def clear(self):
        self.active = False
        js_string = "import('/js/main.js').then(module => {module.removeLayer('" + self.map_id + "')});"
        self.mapbox.view.page().runJavaScript(js_string)

    def update(self):
        print("Updating image layer")

    def set_data(self,
                 x=None,
                 y=None,
                 z=None,
                 image_file=None,
                 legend_title="",
                 cmin=None,
                 cmax=None,
                 cstep=None,
                 decimals=None,
                 crs=None,
                 colormap=None):
        
        self.data = {}
        self.data["x"] = x
        self.data["y"] = y
        self.data["z"] = z
        self.data["image_file"] = image_file
        self.data["legend_title"] = legend_title
        self.data["cmin"] = cmin
        self.data["cmax"] = cmax
        self.data["cstep"] = cstep
        self.data["decimals"] = decimals
        self.data["crs"] = crs
        self.data["colormap"] = colormap

        cmap = colormap

        if not crs:
            src_crs = "EPSG:4326"
        else:
            src_crs = "EPSG:" + crs.epsg

        # Web Mercator
        dst_crs = 'EPSG:3857'

        if image_file:

            with rasterio.open(image_file) as src:
                transform, width, height = calculate_default_transform(
                    src.crs, dst_crs, src.width, src.height, *src.bounds)
                kwargs = src.meta.copy()
                kwargs.update({
                    'crs': dst_crs,
                    'transform': transform,
                    'width': width,
                    'height': height
                })
                # bnds = src.bounds

                mem_file = MemoryFile()
                with mem_file.open(**kwargs) as dst:
                    for i in range(1, src.count + 1):
                        reproject(
                            source=rasterio.band(src, i),
                            destination=rasterio.band(dst, i),
                            src_transform=src.transform,
                            src_crs=src.crs,
                            dst_transform=transform,
                            dst_crs=dst_crs,
                            resampling=Resampling.nearest)

                    band1 = dst.read(1)

            new_bounds = transform_bounds(dst_crs, src_crs,
                                          dst.bounds[0],
                                          dst.bounds[1],
                                          dst.bounds[2],
                                          dst.bounds[3])
            isn = np.where(band1 < 0.001)
            band1[isn] = np.nan

            band1 = np.flipud(band1)
            cminimum = np.nanmin(band1)
            cmaximum = np.nanmax(band1)

            norm = matplotlib.colors.Normalize(vmin=cminimum, vmax=cmaximum)
            vnorm = norm(band1)

            cmap = cm.get_cmap(colormap)
            im = Image.fromarray(np.uint8(cmap(vnorm) * 255))

            overlay_file = "./overlays/" + self.file_name
            im.save(os.path.join(self.mapbox.server_path, overlay_file))

            # Bounds
            bounds = [[new_bounds[0], new_bounds[2]], [new_bounds[3], new_bounds[1]]]
            bounds_string = "[[" + str(bounds[0][0]) + "," + str(bounds[0][1]) + "],[" + str(bounds[1][0]) + "," + str(bounds[1][1]) + "]]"

        else:

            # Color scale
            if self.color_scale_auto:
                if self.color_scale_symmetric:
                    if self.color_scale_symmetric_side == "min":
                        cmin = np.nanmin(z)
                        cmax = -cmin
                    elif self.color_scale_symmetric_side == "max":
                        cmax = np.nanmax(z)
                        cmin = -cmax
                    else:
                        cmx = max(abs(np.nanmin(z)), abs(np.nanmax(z)))
                        cmin = -cmx
                        cmax = cmx
                else:
                    cmin = np.nanmin(z)
                    cmax = np.nanmax(z)
            else:    
                cmin = self.color_scale_cmin
                cmax = self.color_scale_cmax

            ls = LightSource(azdeg=315, altdeg=30)

#            cmap = settings.color_map_earth
#            cmap = matplotlib.pyplot.get_cmap('gist_earth')
            dx = (x[1] - x[0]) / 2
            dy = (y[1] - y[0]) / 2
            rgb = ls.shade(np.flipud(z), cmap,
                           vmin=cmin,
                           vmax=cmax,
                           dx=dx * 50000,
                           dy=dy * 50000,
                           vert_exag=10.0,
                           blend_mode="soft")
            rgb = rgb * 255
            rgb = rgb.astype(np.uint8)

            # First create rasterio image
            res = (x[-1] - x[0]) / (x.size - 1)
            transform = Affine.translation(x[0] - res / 2, y[-1] + res / 2) * Affine.scale(res, -res)
            mem_file1 = MemoryFile()
            src = rasterio.open(
                mem_file1,
                'w',
                driver='GTiff',
                height=y.size,
                width=x.size,
                count=4,
                dtype=rgb.dtype,
                crs=src_crs,
                transform=transform,
            )
            src.write(rgb[:,:,0], 1)
            src.write(rgb[:,:,1], 2)
            src.write(rgb[:,:,2], 3)
            src.write(rgb[:,:,3], 4)

            # Let's convert it to web mercator
            transform, width, height = calculate_default_transform(
                src.crs, dst_crs, src.width, src.height, *src.bounds)
            kwargs = src.meta.copy()
            kwargs.update({
                'crs': dst_crs,
                'transform': transform,
                'width': width,
                'height': height
            })
            # bnds = src.bounds

            mem_file2 = MemoryFile()
            with mem_file2.open(**kwargs) as dst:
                for i in range(1, src.count + 1):
                    reproject(
                        source=rasterio.band(src, i),
                        destination=rasterio.band(dst, i),
                        src_transform=src.transform,
                        src_crs=src.crs,
                        dst_transform=transform,
                        dst_crs=dst_crs,
                        resampling=Resampling.nearest)

                bounds = transform_bounds(dst_crs, src_crs,
                                          dst.bounds[0],
                                          dst.bounds[1],
                                          dst.bounds[2],
                                          dst.bounds[3])

                rgb = np.empty([dst.height, dst.width, 4], dtype=np.uint8)
                rgb[:,:,0] = dst.read(1)
                rgb[:,:,1] = dst.read(2)
                rgb[:,:,2] = dst.read(3)
                rgb[:,:,3] = dst.read(4)
                im = Image.fromarray(rgb, "RGBA")

                overlay_file = "./overlays/" + self.file_name
                im.save(os.path.join(self.mapbox.server_path, "overlays", self.file_name))

                # Bounds
                bounds_string = "[[" + str(bounds[0]) + "," + str(bounds[2]) + "],[" + str(bounds[1]) + "," + str(bounds[3]) + "]]"


                # Delete old legend files
                for file_name in glob.glob(os.path.join(self.mapbox.server_path, "overlays", self.map_id + ".legend.*.png")):
                    try:
                        os.remove(file_name)
                    except:
                        pass

                # add random integer string to legend file to force reload                
                # create string with random integer between 1 and 1,000,000
                rstring = str(np.random.randint(1, 1000000))
                legend_file = self.map_id + ".legend." + rstring + ".png"
                cm2png(cmap,
                    file_name = os.path.join(self.mapbox.server_path, "overlays", legend_file),
                    orientation="vertical",
                    vmin=cmin,
                    vmax=cmax)

        # Legend
        clrbar = ColorBar(colormap=colormap, legend_title=legend_title)
        clrbar.make(cmin, cmax, cstep=cstep, decimals=decimals)
        clrmap_string = clrbar.to_json()

        clrmap_string = "'./overlays/" + legend_file + "'"

        if self.new:
            js_string = "import('/js/image_layer.js').then(module => {module.addLayer('" + overlay_file + "','" + self.map_id + "'," + bounds_string + "," + clrmap_string + ")});"
            self.mapbox.view.page().runJavaScript(js_string)
        else:
            js_string = "import('/js/image_layer.js').then(module => {module.updateLayer('" + overlay_file + "','" + self.map_id + "'," + bounds_string + "," + clrmap_string + ")});"
            self.mapbox.view.page().runJavaScript(js_string)
        self.mapbox.runjs("/js/image_layer.js", "setOpacity", arglist=[self.map_id, self.opacity, self.side])

        self.new = False

    def redraw(self):
        if self.data:
            self.new = True
            self.set_data(x=self.data["x"],
                          y=self.data["y"],
                          z=self.data["z"],
                          image_file=self.data["image_file"],
                          legend_title=self.data["legend_title"],
                          cmin=self.data["cmin"],
                          cmax=self.data["cmax"],
                          cstep=self.data["cstep"],
                          decimals=self.data["decimals"],
                          crs=self.data["crs"],
                          colormap=self.data["colormap"])
        if not self.get_visibility():
            self.set_visibility(False)
        



import traceback

class DependencyCheck:
    def __init__(self, variable_group):
        self.variable       = ""
        self.variable_group = variable_group
        self.operator       = "eq"
        self.value          = None

class Dependency:
    def __init__(self):
        self.action         = "enable"
        self.checkfor       = "all"
        self.checks         = []

    def get(self):
        try:
            getvar = self.gui.getvar
            if self.checkfor == "all":
                okay = True
                for check in self.checks:
                    name = check.variable
                    group = check.variable_group
                    value = getvar(group, name)
                    if type(check.value) == dict:
                        check_value = getvar(check.value.variable_group, check.value.variable)
                    else:
                        check_value = check.value
                        if type(value) == int:
                            check_value = int(check_value)
                        elif type(value) == float:
                            check_value = float(check_value)
                    # if self.check_variables(name=name, group=group):
                    if check.operator == "eq":
                        if value != check_value:
                            okay = False
                            break
                    elif check.operator == "ne":
                        if value == check_value:
                            okay = False
                            break
                    elif check.operator == "gt":
                        if value <= check_value:
                            okay = False
                            break
                    elif check.operator == "ge":
                        if value < check_value:
                            okay = False
                            break
                    elif check.operator == "lt":
                        if value >= check_value:
                            okay = False
                            break
                    elif check.operator == "le":
                        if value > check_value:
                            okay = False
                            break
            elif self.checkfor == "any":
                okay = False
                for check in self.checks:
                    name = check.variable
                    group = check.variable_group
                    value = getvar(group, name)
                    if type(check.value) == dict:
                        check_value = getvar(check.value.variable_group, check.value.variable)
                    else:
                        check_value = check.value
                        if type(value) == int:
                            check_value = int(check_value)
                        elif type(value) == float:
                            check_value = float(check_value)
                    # if self.check_variables(name=name, group=group):
                    if check.operator == "eq":
                        if value == check_value:
                            okay = True
                            break
                    elif check.operator == "ne":
                        if value != check_value:
                            okay = True
                            break
                    elif check.operator == "gt":
                        if value > check_value:
                            okay = True
                            break
                    elif check.operator == "ge":
                        if value >= check_value:
                            okay = True
                            break
                    elif check.operator == "lt":
                        if value < check_value:
                            okay = True
                            break
                    elif check.operator == "le":
                        if value <= check_value:
                            okay = True
                            break
            elif self.checkfor == "none":
                okay = True
                for check in self.checks:
                    name = check.variable
                    group = check.variable_group
                    value = getvar(group, name)
                    if type(check.value) == dict:
                        check_value = getvar(check.value.variable_group, check.value.variable)
                    else:
                        check_value = check.value
                        if type(value) == int:
                            check_value = int(check_value)
                        elif type(value) == float:
                            check_value = float(check_value)
                    if check.operator == "eq":
                        if value == check_value:
                            okay = False
                            break
                    elif check.operator == "ne":
                        if value != check_value:
                            okay = False
                            break
                    elif check.operator == "gt":
                        if value > check_value:
                            okay = False
                            break
                    elif check.operator == "ge":
                        if value >= check_value:
                            okay = False
                            break
                    elif check.operator == "lt":
                        if value < check_value:
                            okay = False
                            break
                    elif check.operator == "le":
                        if value <= check_value:
                            okay = False
                            break
            return okay

        except:
            traceback.print_exc()
            return False
        


import os

from .layer import Layer
# from geopandas import GeoDataFrame

class RasterTileLayer(Layer):
    def __init__(self, mapbox, id, map_id, **kwargs):
        super().__init__(mapbox, id, map_id, **kwargs)

        pass

    def set_data(self,
                 data):
        self.data = data
        # check if the string already contains the x, y, z part
        if "{x}" in data:
            url = data
        else:
            url = data + "/{z}/{x}/{y}.png"
        # url = "https://tile.openstreetmap.org/{z}/{x}/{y}.png"    
        # Add new layer        
        self.mapbox.runjs("./js/raster_tile_layer.js", "addLayer", arglist=[self.map_id,
                                                                    url])
    def redraw(self):
        # if isinstance(self.data, GeoDataFrame):
        #     self.set_data(self.data)
        if not self.get_visibility():
            self.set_visibility(False)

    def activate(self):
        pass
        # self.mapbox.runjs("./js/line_layer.js", "setPaintProperties", arglist=[self.map_id,
        #                                                                        self.line_color,
        #                                                                        self.line_width,
        #                                                                        self.line_opacity,
        #                                                                        self.fill_color,
        #                                                                        self.fill_opacity,
        #                                                                        self.circle_radius])
  
    def deactivate(self):
        pass
        # self.mapbox.runjs("./js/line_layer.js", "setPaintProperties", arglist=[self.map_id,
        #                                                                        self.line_color_inactive,
        #                                                                        self.line_width_inactive,
        #                                                                        self.line_opacity_inactive,
        #                                                                        self.fill_color_inactive,
        #                                                                        self.fill_opacity_inactive,
        #                                                                        self.circle_radius_inactive])

    # def set_visibility(self, true_or_false):
    #     if true_or_false:
    #         self.mapbox.runjs("/js/main.js", "showLayer", arglist=[self.map_id + ".line"])
    #         self.mapbox.runjs("/js/main.js", "showLayer", arglist=[self.map_id + ".circle"])
    #     else:
    #         self.mapbox.runjs("/js/main.js", "hideLayer", arglist=[self.map_id + ".line"])
    #         self.mapbox.runjs("/js/main.js", "hideLayer", arglist=[self.map_id + ".circle"])




from .layer import Layer
from geopandas import GeoDataFrame
from pyogrio import read_dataframe

class MarkerLayer(Layer):
    def __init__(self, mapbox, id, map_id, **kwargs):
        super().__init__(mapbox, id, map_id, **kwargs)
        pass

    def set_data(self, data):
        """Set data for layer"""
        # Make sure this is not an empty GeoDataFrame
        if isinstance(data, GeoDataFrame):
            # Data is GeoDataFrame
            if len(data) == 0:
                return
            if data.crs != 4326:
                data = data.to_crs(4326)
        else:
            # Read geodataframe from shape file
            data = read_dataframe(data)

        self.data = data    

        # Over-ride properties in rows
        if self.icon_file:
            # All markers get the same icon
            data["icon_url"] = self.icon_file
        if self.click_popup_width:
            data["click_popup_width"] = self.click_popup_width
        data["icon_size"] = self.icon_size

        # Add relative path to icon_file 
        # Check if there is a column in the dataframe data called "icon_file"
        if "icon_url" in data.columns:
            # Loop through all rows and add the relative path "./icons/" to the icon_file
            for i, row in data.iterrows():
                if not row["icon_url"].startswith("http"):
                    data.at[i, "icon_url"] = "./icons/" + row["icon_url"]
        else:
            # Use default icon for all markers
            data["icon_url"] = "./icons/mapbox-marker-icon-20px-" + self.marker_color + ".png"

        if self.click_popup_width:
            wdtstr = str(self.click_popup_width) + "px"
        else:
            wdtstr = "100%"
        if self.click_popup_height:
            hgtstr = str(self.click_popup_height) + "px"
        else:
            hgtstr = "100%"
                
        # Check if there is a column in the geodataframe called "click_html"
        if self.click_property:
            # Add click_html column to data
            data["click_html"] = ""
            # Loop through all rows
            for i, row in data.iterrows():
                # Loop through all rows and check if click_html ends with .html and does not start with http
                # If so, replace with relative path in server folder
                if row[self.click_property].endswith(".html") and not row[self.click_property].startswith("http"):
                    # Must be a local html file
                    data.at[i, "click_html"] = '<div><iframe width="' + wdtstr + '" height="' + hgtstr + '" src="' + './overlays/' + row[self.click_property] + '"></iframe></div>'
                elif row[self.click_property].startswith("http"):
                    # Must be an external url (this often does not work and results in an error: Refused to display in a frame because it set 'X-Frame-Options' to 'sameorigin)
                    data.at[i, "click_html"] = '<div><iframe width="' + wdtstr + '" height="' + hgtstr + '" src="' + row[self.click_property] + '"></iframe></div>'
                else:
                    # Get click_html from click_property
                    data.at[i, "click_html"] = row[self.click_property]    

        # Do the same for hover_property and hover_html
        if self.hover_property:
            data["hover_html"] = ""
            for i, row in data.iterrows():
                if row[self.hover_property].endswith(".html") and not row[self.hover_property].startswith("http"):
                    data.at[i, "hover_html"] = '<div><iframe src="' + './overlays/' + row[self.hover_property] + '"></iframe></div>'
                elif row[self.hover_property].startswith("http"):
                    data.at[i, "hover_html"] = '<div><iframe src="' + row[self.hover_property] + '"></iframe></div>'
                else:
                    data.at[i, "hover_html"] = row[self.hover_property]
         
        # Add new layer
        self.mapbox.runjs(
            "./js/marker_layer.js",
            "addLayer",
            arglist=[
                self.map_id,
                data
            ],
        )
        pass

    # TODO: Implement activate and deactivate

    # 
    # def activate(self):
    #     self.show()
    #     self.mapbox.runjs(
    #         "./js/geojson_layer_circle.js",
    #         "setPaintProperties",
    #         arglist=[
    #             self.map_id,
    #             self.line_color,
    #             self.line_width,
    #             self.line_opacity,
    #             self.fill_color,
    #             self.fill_opacity,
    #             self.circle_radius,
    #         ],
    #     )

    # def deactivate(self):
    #     self.mapbox.runjs(
    #         "./js/geojson_layer_circle.js",
    #         "setPaintProperties",
    #         arglist=[
    #             self.map_id,
    #             self.line_color_inactive,
    #             self.line_width_inactive,
    #             self.line_opacity_inactive,
    #             self.fill_color_inactive,
    #             self.fill_opacity_inactive,
    #             self.circle_radius_inactive,
    #         ],
    #     )

    def redraw(self):
        if isinstance(self.data, GeoDataFrame):
            self.set_data(self.data)
        if not self.get_visibility():
            self.set_visibility(False)



from http.server import HTTPServer as BaseHTTPServer, SimpleHTTPRequestHandler
import threading
import os
import time
import urllib

# Custom Exception Class
class MyException(Exception):
    pass
 
# Custom Thread Class
class ServerThread(threading.Thread):
    def __init__(self, server_path, server_port):
        super().__init__(daemon=True)
        self.server_path = server_path
        self.server_port = server_port
     
  # Function that raises the custom exception
    def run_server(self):
        print("Server path : " + self.server_path)
        httpd = HTTPServer(self.server_path, ("", self.server_port))
        httpd.serve_forever()
        name = threading.current_thread().name
        raise MyException("An error in thread " + name)
 
    def run(self):       
        # Variable that stores the exception, if raised by run_server
        self.exc = None           
        try:
            self.run_server()
        except BaseException as e:
            self.exc = e
       
    # def join(self):
    #     threading.Thread.join(self)
    #     # Since join() returns in caller thread
    #     # we re-raise the caught exception
    #     # if any was caught
    #     if self.exc:
    #         raise self.exc
        
class HTTPHandler(SimpleHTTPRequestHandler):
    """This handler uses server.base_path instead of always using os.getcwd()"""
    def translate_path(self, path):
        path = SimpleHTTPRequestHandler.translate_path(self, path)
        relpath = os.path.relpath(path, os.getcwd())
        fullpath = os.path.join(self.server.base_path, relpath)
        return fullpath

class HTTPServer(BaseHTTPServer):
    """The main server, you pass in base_path which is the path you want to serve requests from"""
    def __init__(self, base_path, server_address, RequestHandlerClass=HTTPHandler):
        self.base_path = base_path
        BaseHTTPServer.__init__(self, server_address, RequestHandlerClass)

# def run_server(server_path, server_port):
#     print("Server path : " + server_path)
#     httpd = HTTPServer(server_path, ("", server_port))
#     httpd.serve_forever()

def run_node_server(server_path, server_port):
    print("Node server path : " + server_path)
    os.chdir(server_path)
    os.system("run_node_server.bat")

def start_server(server_path, port=3000, node=False):
    if node:
        thr = threading.Thread(target=run_node_server, args=(server_path, port), daemon=True)
        thr.start()
        while True:
            try:        
                print("Finding server ...")        
                urllib.request.urlcleanup()
                url = "http://localhost:" + str(port)
                request = urllib.request.urlopen(url)
                request.close()
                print("Found server running at port " + str(port) + " ...")        
                break        
            except:
                print("Waiting for server ...")
                time.sleep(1)
        return thr       
    else:
#        thr = threading.Thread(target=run_server, args=(server_path, port), daemon=True)
        thr = ServerThread(server_path, port)
        thr.start()
        # # Exception handled in Caller thread
        # try:
        #     thr.join()
        # except Exception as e:
        #     print("Exception Handled in Main, Details of the Exception:", e)
        while True:
            try:        
                print("Finding server ...")        
                urllib.request.urlcleanup()
                url = "http://localhost:" + str(port)
                request = urllib.request.urlopen(url)
                request.close()
                print("Found server running at port " + str(port) + " ...")        
                break        
            except:
                print("Waiting for server ...")
                time.sleep(1)
        return thr       



import math
import geopandas as gpd
import matplotlib.colors as mcolors
import shapely
import json
from shapely.geometry import Polygon
from shapely.ops import transform
import pyproj

from .layer import Layer


class DrawLayer(Layer):
    def __init__(
        self,
        mapbox,
        id,
        map_id,
        shape="polygon",
        create=None,
        modify=None,
        select=None,
        deselect=None,
        add=None,
        rotate=True,
        polygon_line_color="dodgerblue",
        polygon_line_width=2,
        polygon_line_style="-",
        polygon_line_opacity=1.0,
        polygon_fill_color="dodgerblue",
        polygon_fill_opacity=0.5,
        polyline_line_color="limegreen",
        polyline_line_width=2,
        polyline_line_style="-",
        polyline_line_opacity=1.0,
        circle_line_color="black",
        circle_line_opacity=1.0,
        circle_fill_color="orangered",
        circle_fill_opacity=0.5,
        circle_radius=4,
    ):
        super().__init__(mapbox, id, map_id)

        self.active = False
        self.type = "draw"
        self.shape = shape
        self.mode = (
            "active"  # Draw layers can have three modes: active, inactive, invisible
        )
        self.gdf = gpd.GeoDataFrame()
        self.create   = create
        self.modify   = modify
        self.select   = select
        self.deselect = deselect
        self.add      = add

        # Get Hex values for colors
        self.paint_props = {}
        self.paint_props["polygon_line_color"] = mcolors.to_hex(polygon_line_color)
        self.paint_props["polygon_line_width"] = polygon_line_width
        self.paint_props["polygon_line_style"] = polygon_line_style
        self.paint_props["polygon_line_opacity"] = polygon_line_opacity
        self.paint_props["polygon_fill_color"] = mcolors.to_hex(polygon_fill_color)
        self.paint_props["polygon_fill_opacity"] = polygon_fill_opacity
        self.paint_props["polyline_line_color"] = mcolors.to_hex(polyline_line_color)
        self.paint_props["polyline_line_width"] = polyline_line_width
        self.paint_props["polyline_line_style"] = polyline_line_style
        self.paint_props["polyline_line_opacity"] = polyline_line_opacity
        self.paint_props["circle_line_color"] = mcolors.to_hex(circle_line_color)
        self.paint_props["circle_line_opacity"] = circle_line_opacity
        self.paint_props["circle_fill_color"] = mcolors.to_hex(circle_fill_color)
        self.paint_props["circle_fill_opacity"] = circle_fill_opacity
        self.paint_props["circle_radius"] = circle_radius
        self.paint_props["rotate"] = rotate

        # Add this layer
        self.mapbox.runjs(
            "./js/draw_layer.js",
            "addLayer",
            arglist=[self.map_id, "active", self.paint_props, self.shape],
        )

    def set_mode(self, mode):
        """Set mode of layer. Can be active, inactive or invisible."""
        self.mode = mode
        if mode == "active":
            self.active = True
            self.visible = True
        elif mode == "inactive":
            self.active = False
            self.visible = True
        else:
            self.active = False
            self.visible = False
        self.mapbox.runjs("./js/draw_layer.js", "setLayerMode", arglist=[self.map_id, mode])

    def activate(self):
        """Activate the draw layer so it can be edited by the user."""
        self.active = True
        self.mapbox.runjs("./js/draw_layer.js", "setLayerMode", arglist=[self.map_id, "active"])

    def deactivate(self):
        """Deactivate the draw layer so it can not be edited by the user."""
        self.active = False
        self.mapbox.runjs("./js/draw_layer.js", "setLayerMode", arglist=[self.map_id, "inactive"])

    # def show(self):
    #     self.visible = True
    #     if self.active:
    #         self.mapbox.runjs("./js/draw_layer.js", "setLayerMode", arglist=[self.map_id, "active"])
    #     else:
    #         self.mapbox.runjs("./js/draw_layer.js", "setLayerMode", arglist=[self.map_id, "inactive"])    

    # def hide(self):
    #     self.visible = False
    #     self.mapbox.runjs("./js/draw_layer.js", "setLayerMode", arglist=[self.map_id, "invisible"])

    # def set_mode(self, mode):
    #     self.mode = mode
    #     self.mapbox.runjs("./js/draw_layer.js", "setLayerMode", arglist=[self.map_id, mode])

    def set_visibility(self, true_or_false):
        if true_or_false:
            if self.active:
                self.mapbox.runjs("./js/draw_layer.js", "setLayerMode", arglist=[self.map_id, "active"])
            else:
                self.mapbox.runjs("./js/draw_layer.js", "setLayerMode", arglist=[self.map_id, "inactive"])    
        else:
            # Make layer invisible
            self.mapbox.runjs("./js/draw_layer.js", "setLayerMode", arglist=[self.map_id, "invisible"])    

    def set_data(self, gdf):
        """Clear the draw layer and add data. Data must be a GeoDataFrame."""
        self.clear()
        self.add_feature(gdf)

    def add_feature(self, gdf):
        """Add data to draw layer. Data must be a GeoDataFrame."""
        # Loop through features
        if len(gdf) == 0:
            return
    
        # Append to self.gdf (which doesn't have to have CRS=4326)
        if not self.gdf.empty:
            self.gdf = self.gdf.append(gdf.to_crs(self.gdf.crs), ignore_index=True)
        else:
            self.gdf = gdf
        
        for index, row in gdf.to_crs(4326).iterrows():
            # Add feature to draw layer, which has CRS=4326
            gdf = gpd.GeoDataFrame(geometry=[row["geometry"]], crs=4326)
            self.mapbox.runjs("./js/draw_layer.js", "addFeature", arglist=[gdf, self.map_id])

    def add_rectangle(self, x0, y0, lenx, leny, rotation):
        """Add rectangle to draw layer."""
        x = [x0]
        y = [y0]
        x.append(x[0] + lenx * math.cos(math.pi*rotation/180))
        y.append(y[0] + lenx * math.sin(math.pi*rotation/180))
        x.append(x[1] + leny * math.cos(math.pi*(rotation + 90.0)/180))
        y.append(y[1] + leny * math.sin(math.pi*(rotation + 90.0)/180))
        x.append(x[2] + lenx * math.cos(math.pi*(rotation + 180.0)/180))
        y.append(y[2] + lenx * math.sin(math.pi*(rotation + 180.0)/180))
        x.append(x0)
        y.append(y0)
        polygon_crs = Polygon(zip(x, y))
        # Transform to WGS 84
        project = pyproj.Transformer.from_crs(
            self.mapbox.crs, pyproj.CRS(4326), always_xy=True
        ).transform
        polygon_wgs84 = transform(project, polygon_crs)
        gdf = gpd.GeoDataFrame(geometry=[polygon_wgs84], crs=4326)
        self._x0 = x0
        self._y0 = y0
        self._dx = lenx
        self._dy = leny
        self._rotation = math.pi*rotation/180
        self.add_feature(gdf)

    def draw(self):
        """Activate drawing mode."""
        if self.shape == "polygon":
            self.mapbox.runjs("./js/draw_layer.js", "drawPolygon", arglist=[self.map_id])
        elif self.shape == "polyline":
            self.mapbox.runjs("./js/draw_layer.js", "drawPolyline", arglist=[self.map_id])
        elif self.shape == "rectangle":
            self.mapbox.runjs("./js/draw_layer.js", "drawRectangle", arglist=[self.map_id])

    def set_gdf(self, feature_collection, compute_geometry=True):
        # Called after a feature has been drawn or added. A new GeoDataFrame is created for this layer.
        for feature in feature_collection["features"]:
            feature["properties"]["id"] = feature["id"]
        gdf = gpd.GeoDataFrame.from_features(feature_collection, crs=4326).to_crs(
            self.mapbox.crs
        )
        if self.shape == "rectangle":
            if compute_geometry:
                x0, y0, dx, dy, rotation = get_rectangle_geometry(gdf["geometry"])
                # Add columns with geometry info
                gdf["x0"] = x0
                gdf["y0"] = y0
                gdf["dx"] = dx
                gdf["dy"] = dy
                gdf["rotation"] = rotation
            else:
                gdf["x0"] = self._x0
                gdf["y0"] = self._y0
                gdf["dx"] = self._dx
                gdf["dy"] = self._dy
                gdf["rotation"] = self._rotation
        self.gdf = gdf

    def get_feature_index(self, feature_id):
        """Get a features index by ID in the GeoDataFrame."""
        feature_index = None
        if len(self.gdf) > 0:
            indx = self.gdf.index[self.gdf["id"] == feature_id].tolist()
            if indx:
                feature_index = indx[0]
        if feature_index is None:
            print("Could not find feature ...")
        return feature_index

    def get_feature_id(self, feature_index):
        """Get a features ID by index in the GeoDataFrame."""
        feature_id = None
        if len(self.gdf) > 0 and len(self.gdf) >= feature_index + 1:
            feature_id = self.gdf.loc[feature_index, "id"]
        return feature_id

    def feature_drawn(self, feature_collection, feature_id):
        # Called after a feature has been drawn by the user
        self.set_gdf(feature_collection)
        if self.create:
            feature_index = self.get_feature_index(feature_id)
            self.create(self.gdf, feature_index, feature_id)
        if self.shape == "rectangle" and not self.mapbox.crs.is_geographic:
            feature_index = self.get_feature_index(feature_id)
            x0 = self.gdf.loc[feature_index, "x0"]  
            y0 = self.gdf.loc[feature_index, "y0"]
            lenx = self.gdf.loc[feature_index, "dx"]
            leny = self.gdf.loc[feature_index, "dy"]
            rotation = self.gdf.loc[feature_index, "rotation"]*180/math.pi
            self.delete_feature(feature_id)
            self.add_rectangle(x0, y0, lenx, leny, rotation)

    def feature_added(self, feature_collection, feature_id):
        # Called after a feature has been added by the application
        self.set_gdf(feature_collection, compute_geometry=False)
#        self.set_visibility(True)
#        self.set_mode(self.mode)
        if self.add:
            feature_index = self.get_feature_index(feature_id)
            self.add(self.gdf, feature_index, feature_id)

    def feature_modified(self, feature_collection, feature_id):
        # Called after a feature has been modified by the user
        self.set_gdf(feature_collection)
        if self.modify:
            feature_index = self.get_feature_index(feature_id)
            self.modify(self.gdf, feature_index, feature_id)

    def feature_selected(self, feature_collection, feature_id):
        # Called after a feature has been selected by the user
        if self.select:
            feature_index = self.get_feature_index(feature_id)
            self.select(feature_index)

    def feature_deselected(self):
        # Called after a feature has been de-selected by the user
        if self.deselect:
            self.deselect()

    def activate_feature(self, feature_id_or_index):
        """Activate a feature by ID or index so it can be edited by the user."""
        if isinstance(feature_id_or_index, int):
            # It's an index
            index = feature_id_or_index
            feature_id = self.get_feature_id(index)
        else:
            feature_id = feature_id_or_index    
        if self.mode != "active":
            self.mode = "active"
            self.mapbox.runjs(
                "./js/draw_layer.js", "setLayerMode", arglist=[self.map_id, self.mode]
            )
        self.mapbox.runjs("./js/draw_layer.js", "activateFeature", arglist=[feature_id])

    def set_feature_geometry(self, feature_id, geom):        
        self.mapbox.runjs("./js/draw_layer.js", "setFeatureGeometry", arglist=[self.map_id, feature_id, geom])

    def delete_feature(self, feature_id_or_index):
        """Delete a feature by ID or index"""
        if isinstance(feature_id_or_index, int):
            # It's an index
            index = feature_id_or_index
            feature_id = self.get_feature_id(index)
            # Remove from gdf
            self.gdf = self.gdf.drop(index)
            if len(self.gdf) > 0:
                self.gdf = self.gdf.reset_index(drop=True)
            else:
                self.gdf = gpd.GeoDataFrame()
            # Remove from map
            self.mapbox.runjs("./js/draw_layer.js", "deleteFeature", arglist=[feature_id])
        else:    
            feature_id = feature_id_or_index
            # Remove from gdf
            for index, row in self.gdf.iterrows():
                if row["id"] == feature_id:
                    self.gdf = self.gdf.drop(index)
                    if len(self.gdf) > 0:
                        self.gdf = self.gdf.reset_index(drop=True)
                    else:
                        self.gdf = gpd.GeoDataFrame()    
                    break
            self.mapbox.runjs("./js/draw_layer.js", "deleteFeature", arglist=[feature_id])
        return self.gdf    

    def delete_from_map(self):
        """Delete the draw layer from the map."""
        self.active = False
        self.mapbox.runjs("./js/draw_layer.js", "deleteLayer", arglist=[self.map_id])
        self.gdf = gpd.GeoDataFrame()

    def clear(self):
        """Clear the draw layer."""
        for index, row in self.gdf.iterrows():
            self.mapbox.runjs("./js/draw_layer.js", "deleteFeature", arglist=[row["id"]])
        self.gdf = gpd.GeoDataFrame()

    def get_gdf(self, id=None):
        """Return the GeoDataFrame for this layer. If id is specified, return only the row with that id."""
        if id:
            for index, row in self.gdf.iterrows():
                if row["id"] == id:
                    return self.gdf[index]
        else:
            return self.gdf

    def redraw(self):
        # Add this layer
        self.mapbox.runjs(
            "./js/draw_layer.js",
            "addLayer",
            arglist=[self.map_id, "active", self.paint_props, self.shape],
        )
        self.add_feature(self.gdf)
        self.set_mode(self.mode)
        if not self.get_visibility():
            self.set_visibility(False)

def get_rectangle_geometry(geoms):
    x0 = []
    y0 = []
    dx = []
    dy = []
    rotation = []
    for geom in geoms:
        xx, yy = geom.exterior.coords.xy
        x0.append(float(xx[0]))
        y0.append(float(yy[0]))
        lenx = math.sqrt(float(xx[1] - xx[0])**2 + float(yy[1] - yy[0])**2)
        leny = math.sqrt(float(xx[2] - xx[1])**2 + float(yy[2] - yy[1])**2)
        dx.append(lenx)
        dy.append(leny)
        rot = float((math.atan2(yy[1] - yy[0], xx[1] - xx[0])))
        if abs(rot*180/math.pi) < 1.0:
            rot = 0.0
        rotation.append(rot)

    return x0, y0, dx, dy, rotation



import importlib

from guitares.dependencies import Dependency, DependencyCheck

class Text:
    def __init__(self, variable_group):
        self.variable       = ""
        self.variable_group = variable_group

class Menu:
    def __init__(self, dct, parent):
        self.style = None
        self.menus = []
        self.dependencies = []
        self.parent = parent
        self.text = ""
        self.option = None
        self.checkable = False
        self.separator = False
        self.dependencies = []
        self.visible = True
        self.enable = True
        self.id = ""
        self.variable_group = parent.variable_group
        self.module = parent.module
        self.method = ""
        self.callback = None
        self.gui = parent.gui
        self.getvar = parent.gui.getvar
        self.setvar = parent.gui.setvar
        self.icon = ""

        # Now update element attributes based on dict

        if "variable_group" in dct:
            self.variable_group = dct["variable_group"]
        if "id" in dct:
            self.id = dct["id"]
        if "option" in dct:
            self.option = dct["option"]
        if "checkable" in dct:
            self.checkable = dct["checkable"]
        if "separator" in dct:
            self.separator = dct["separator"]
        if "module" in dct:
            if type(dct["module"]) == str:
                try:
                    self.module = importlib.import_module(dct["module"])
                except:
                    print("Error! Could not import module " + dct["module"])
        if "method" in dct:
            self.method = dct["method"]
        if self.method and self.module:
            if hasattr(self.module, self.method):
                self.callback = getattr(self.module, self.method)
            else:
                print("Error! Could not find method " + self.method)
        if "text" in dct:
            if type(dct["text"]) == dict:
                self.text = Text(self.variable_group)
                if "variable" in dct:
                    self.text.variable(dct["variable"])
                if "variable_group" in dct:
                    self.text.variable_group(dct["variable_group"])
            else:
                self.text = dct["text"]

        if "dependency" in dct:
            for dep in dct["dependency"]:
                dependency = Dependency()
                dependency.gui = parent.gui
                if "action" in dep:
                    dependency.action = dep["action"]
                if "checkfor" in dep:
                    dependency.checkfor = dep["checkfor"]
                for check_dct in dep["check"]:
                    check = DependencyCheck(self.variable_group)
                    if "variable" in check_dct:
                        check.variable = check_dct["variable"]
                    if "variable_group" in check_dct:
                        check.variable_group = check_dct["variable_group"]
                    if "operator" in check_dct:
                        check.operator = check_dct["operator"]
                    if "value" in check_dct:
                        check.value = check_dct["value"]
                    dependency.checks.append(check)
                self.dependencies.append(dependency)


    def add(self):
        if self.menus:
            from .pyqt5.menu import Menu
            Menu(self)
        else:
            # End node
            from .pyqt5.menu import Action
            Action(self)

    def set_dependencies(self):
        for dependency in self.dependencies:
            okay = dependency.get()
            if dependency.action == "enable":
                if okay:
                    self.widget.setEnabled(True)
                else:
                    self.widget.setEnabled(False)
            elif dependency.action == "check":
                if okay:
                    self.widget.setChecked(True)
                else:
                    self.widget.setChecked(False)



from PyQt5.QtWidgets import QComboBox
from PyQt5.QtWidgets import QLabel
from PyQt5 import QtCore
import traceback

class PopupMenu(QComboBox):
    def __init__(self, element):
        super().__init__(element.parent.widget)

        self.element = element

        if element.option_string.variable:
            group = element.option_string.variable_group
            name  = element.option_string.variable
            v     = element.getvar(group, name)
            for txt in v:
                self.addItem(txt)
        else:
            for txt in element.option_string.list:
                self.addItem(txt)

        # Adjust list value types
        if element.option_value:
            if not type(element.option_value.variable):
                for i, val in enumerate(element.option_value.list):
                    if element.type == float:
                        element.option_value[i] = float(val)
                    elif element.type == int:
                        element.option_value[i] = int(val)

        if element.text:
            if isinstance(element.text, str):
                txt = element.text
            else:
                txt = self.element.getvar(element.text.variable_group, element.text.variable)    
            label = QLabel(txt, element.parent.widget)
            label.setStyleSheet("background: transparent; border: none")
            self.text_widget = label

        if self.element.tooltip:
            if isinstance(self.element.tooltip, str):
                txt = self.element.tooltip
            else:
                txt = self.element.getvar(self.element.tooltip.variable_group, self.element.tooltip.variable)    
            self.setToolTip(txt)

        self.currentIndexChanged.connect(self.callback)

        self.set_geometry()

        self.execute_callback = True

    def showEvent(self, event):
        # Call when this widget becomes visible. This was commented out for the frame widget. Doing the same here. Not sure why this was needed.
        # self.set_geometry()
        pass

    def set(self):

        self.execute_callback = False

        if self.element.option_string.variable:
            self.add_items()

        if self.element.select == "item":
            if self.element.option_value:
                if self.element.option_value.variable:
                    name  = self.element.option_value.variable
                    group = self.element.option_value.variable_group
                    self.element.option_value.list = self.element.getvar(group, name)

            val = self.element.getvar(self.element.variable_group, self.element.variable)
            if val:
                if self.element.option_value:
                    index = self.element.option_value.list.index(val)
                    self.setCurrentIndex(index)
        else:
            name = self.element.variable
            group = self.element.variable_group
            index = self.element.getvar(group, name)
            self.setCurrentIndex(index)

        if not isinstance(self.element.text, str):
            txt = self.element.getvar(self.element.text.variable_group, self.element.text.variable)
            self.text_widget.setText(txt)

        if not isinstance(self.element.tooltip, str):
            txt = self.element.getvar(self.element.tooltip.variable_group, self.element.tooltip.variable)    
            self.setToolTip(txt)

        self.execute_callback = True

    def callback(self):

        if not self.execute_callback:
            return
        
        i = self.currentIndex()
        if self.element.select == "item":
            newval = self.element.option_value.list[i]
        else:
            newval = i

        name  = self.element.variable
        group = self.element.variable_group
        self.element.setvar(group, name, newval)

        try:
            if self.isEnabled() and self.element.callback:
                self.element.callback(newval, self)
                # Update GUI
            self.element.window.update()
        except:
            traceback.print_exc()

    def set_geometry(self):
        resize_factor = self.element.gui.resize_factor
        x0, y0, wdt, hgt = self.element.get_position()
        self.setGeometry(x0, y0, wdt, hgt)
        if self.element.text:
            label = self.text_widget
            fm = label.fontMetrics()
            wlab = int(fm.size(0, self.element.text).width())
            if self.element.text_position == "above-center" or self.element.text_position == "above":
                label.setAlignment(QtCore.Qt.AlignCenter)
                label.setGeometry(x0, int(y0 - 20 * resize_factor), wdt, int(20 * resize_factor))
            elif self.element.text_position == "above-left":
                label.setAlignment(QtCore.Qt.AlignLeft)
                label.setGeometry(x0, int(y0 - 20 * resize_factor), wlab, int(20 * resize_factor))
            else:
                # Assuming left
                label.setAlignment(QtCore.Qt.AlignRight)
                label.setGeometry(int(x0 - wlab - 3 * resize_factor),
                                  int(y0 + 5 * self.element.gui.resize_factor),
                                  wlab,
                                  int(20 * resize_factor))

    def add_items(self):
        # Delete existing items
        self.clear()
        if self.element.option_string.variable:
            group = self.element.option_string.variable_group
            name  = self.element.option_string.variable
            v     = self.element.getvar(group, name)
            if not v:
                v = [""]
            for itxt, txt in enumerate(v):
                self.insertItem(itxt, txt)
        else:
            for itxt, txt in enumerate(self.element.option_string.list):
                self.insertItem(itxt, txt)



def window_dialog(
    window,
    text,
    title=" ",
    type="warning",
    button_text=None,
    nmax=100,
    cancel=None,
    filter="*.*",
    selected_filter=None,
    path=None,
    file_name=None
):
    if window.gui.framework == "pyqt5":
        from .pyqt5.dialog import dialog

    p = dialog(
        window.widget,
        text,
        title=title,
        type=type,
        button_text=button_text,
        nmax=nmax,
        cancel=cancel,
        filter=filter,
        selected_filter=selected_filter,
        path=path,
        file_name=file_name
    )

    return p



import traceback

class Menu:
    def __init__(self, menu):
        menu.widget = menu.parent.widget.addMenu("&" + menu.text)
        if menu.separator:
            menu.parent.widget.addSeparator()

class Action:
    def __init__(self, menu):
        menu.widget = menu.parent.widget.addAction("&" + menu.text)
        if menu.checkable:
            menu.widget.setCheckable(True)
        if menu.callback:
            menu.widget.triggered.connect(lambda state, f=menu.callback, o=menu.option: self.menu_item_selected(f, o))
        if menu.separator:
            menu.parent.widget.addSeparator()

    def menu_item_selected(self, callback, option):
        try:
            callback(option)
        except:
            traceback.print_exc()



import importlib
from PyQt5 import QtCore
from PyQt5.QtWidgets import QWidget, QVBoxLayout, QStatusBar, QLabel, QFrame

#from guitares.dependencies import Dependency, DependencyCheck

class StatusBar:
    def __init__(self, window, widths):
        sb = QStatusBar()
        window.setStatusBar(sb)
        self.window = window
        self.widget = sb
        self.field  = []
        for w in widths:
            self.field.append({"width": w})

        # Add fields
        ww = window.geometry().width()
        x = 0
        for field in self.field:
            frame = QFrame()
            sb.addWidget(frame)
            label = QLabel("text01", frame)
            field["frame"] = frame
            field["label"] = label
            x = x + int(ww*w) + 5

        self.set_geometry()    

    def set_geometry(self):
        window_width = self.window.geometry().width()
        x = 0
        for field in self.field:
            frame = field["frame"]
            frame.setGeometry(x, 0, int(window_width*field["width"]), 20)
            x = x + int(window_width*field["width"]) + 5
